/////////////////////////////////////////////////////////////////////////////////
//
//  XMLexception: vynimka vyvolavana pri chybach pri parsrovani xml suboru
//  XMLexceptionNotFound: vynimka - polozka nenajdena
//  XMLexceptionInvalid: vynimka - nastala chyba, polozka je neplatna
//  XMLbase: bazova trieda pre XML* triedy.. poskytuje uzitocnu funkcionalitu, ktora sa opakuje
//    pri spracovavani roznych XML-elementov
//  XMLfunc: bazova trieda obalujuca funkcionalitu xml-funkcii
//  XMLfuncInter: trieda obalujuca linearnu interpolaciu definovanu v xml subore
//  XMLfuncPolynome: trieda obalujuca polynomialnu funkciu
//  XML* - nazov napovie
//
/////////////////////////////////////////////////////////////////////////////////
#pragma once

#include <xercesc/dom/DOMNode.hpp>
#include "func.h"
#include "hdrimage.h"
#include "adaptation.h"
#include "featurecolorfilterarray.h"
#include <string>
#include <vector>
#include <sstream>

using std::string;

class XMLfunc;
typedef boost::shared_ptr<XMLfunc> PtrXMLfunc;


class XMLexception
{
	string m_camera;
	string m_version;
	string m_node;
	string m_desc;
public:
	XMLexception (string camera, string version, string node = "", string desc = "")
	{
		m_camera = camera;
		m_version = version;
		m_node = node;
		m_desc = desc;
	}

	string getCamera ()
	{
		return m_camera;
	}
	string getVersion ()
	{
		return m_version;
	}
	string getNode ()
	{
		return m_node;
	}
	string getDesc ()
	{
		return m_desc;
	}

	string get ()
	{
		std::ostringstream oss;
		oss << "<camera_dataset>";
		if (!m_camera.empty() || !m_version.empty())
			oss << "/<camera name=" << m_camera << " version=" << m_version << ">";
		if (0 != m_node.compare (""))
			oss << "/<" << m_node << ">";
		if (0 != m_desc.compare (""))
			oss << ": " << m_desc;
		return oss.str ();
	}
};

class XMLexceptionNotFound: public XMLexception
{
public:
	XMLexceptionNotFound (string camera, string version, string node = "", string desc = ""):
		XMLexception (camera, version, node, desc + " not found")
	{
	}
};

class XMLexceptionInvalid: public XMLexception
{
public:
	XMLexceptionInvalid (string camera, string version, string node = "", string desc = ""):
		XMLexception (camera, version, node, desc + " is invalid")
	{
	}
};
class XMLbase
{
	string m_strCamera;
	string m_strVersion;
	string m_strNode;

public:
	XMLbase (string camera, string version, string node = "")
	{
		m_strCamera = camera;
		m_strVersion = version;
		m_strNode = node;
	}
	bool checkAttribute (xercesc::DOMNode* node, string name, string value);
	xercesc::DOMNode* getChild (std::string name, xercesc::DOMNode* parent);
	xercesc::DOMNode* getChildNotThrowNotFound (std::string name, xercesc::DOMNode* parent);
	bool getAttributeBool (xercesc::DOMNode* node, string name);
	string getAttributeString (xercesc::DOMNode* node, string name);
	float getAttributeFloat (xercesc::DOMNode* node, string name);
	int getAttributeInt (xercesc::DOMNode* node, string name);
	uint getAttributeUInt (xercesc::DOMNode* node, string name);
	PtrXMLfunc getChildFunc (xercesc::DOMNode* node);
	void getGrid (xercesc::DOMNode* node, HDRImage1f& gridX, HDRImage1f& gridY);

	float getFloat  (xercesc::DOMNode* node);
	float2 getFloat2 (xercesc::DOMNode* node);
	float3 getFloat3  (xercesc::DOMNode* node);

	void throwNotFound (std::string object)
	{
		throw XMLexceptionNotFound (m_strCamera, m_strVersion, m_strNode, object);
	}
	void throwInvalid (std::string object)
	{
		throw XMLexceptionInvalid (m_strCamera, m_strVersion, m_strNode, object);
	}
	void throwException (std::string desc)
	{
		throw XMLexception (m_strCamera, m_strVersion, m_strNode, desc);
	}
};



class XMLfunc: public FuncFtoF, protected XMLbase
{
public:
	XMLfunc (string camera, string version):
		XMLbase (camera, version)
	{
	}
};



class XMLfuncInter: public XMLfunc
{
	float2* m_points;
public:
	XMLfuncInter(string camera, string version, xercesc::DOMNode* node);

	float operator() (float x)
	{
		int i;
		for (i = 1; true; i++)
		{
			if (x <= m_points[i].x)
				break;
		}
		float2 low = m_points[i-1];
		float2 high = m_points[i];
		float2 size = high - low;
		float rel = (x - low.x)/size.x; 
		return low.y + size.y*rel;
	}
};




template <uint Degree>
class XMLfuncPolynome: public XMLfunc
{
	float m_coef[Degree+1];
public:
	XMLfuncPolynome(string camera, string version, xercesc::DOMNode* node);

	float operator () (float x)
	{
		float xn= 1.0f;
		float y = 0;
		for (uint i = 0; i <= Degree; i++)
		{
			y += m_coef[i]*xn;
			xn *= x;
		}
		return y;
	}
};


class XMLchromAbberation: protected XMLbase
{
	bool m_enabled;
	PtrXMLfunc m_redcyan, m_blueyellow;
	uint2 m_sizeOfGrid;

public:
	XMLchromAbberation (string camera, string version, xercesc::DOMNode* node);

	bool isEnabled ()
	{
		return m_enabled;
	}

	uint2 getSizeOfGrid ()
	{
		return m_sizeOfGrid;
	}
	
	PtrXMLfunc getFuncRedCyan ()
	{
		return m_redcyan;
	}

	PtrXMLfunc getFuncBlueYellow ()
	{
		return m_blueyellow;
	}
};

typedef boost::shared_ptr<XMLchromAbberation> PtrXMLchromAbberation;


class XMLgeometricDistortion: protected XMLbase
{
	bool m_enabled;
public:
	enum etype {TYPE_NET};
private:
	etype m_type;
	PtrHDRImage1f m_gridX, m_gridY;

	void processRadialDistortion (xercesc::DOMNode* node);
public:
	XMLgeometricDistortion (string camera, string version, xercesc::DOMNode* node);
	
	etype getType ()
	{
		return m_type;
	}
	bool isEnabled ()
	{
		return m_enabled;
	}
	void getGrid (PtrHDRImage1f&  gridX, PtrHDRImage1f& gridY)
	{
		gridX = m_gridX;
		gridY = m_gridY;
	}
};

typedef boost::shared_ptr<XMLgeometricDistortion> PtrXMLgeometricDistortion;

class XMLvignetting: protected XMLbase
{
	bool m_enabled;
	string m_irregFilepath;
	PtrXMLfunc m_regFunc;
public:
	enum type_t {TYPE_REGULAR, TYPE_IRREGULAR};
private:
	type_t m_type;
public:
	XMLvignetting (string camera, string version, xercesc::DOMNode* node);
	bool isEnabled ()
	{
		return m_enabled;
	}
	type_t getType ()
	{
		return m_type;
	}
	string getIrregFilepath ()
	{
		assert (m_type == TYPE_IRREGULAR);
		return m_irregFilepath;
	}
	PtrFuncFtoF getRegFunc ()
	{
		assert (m_type == TYPE_REGULAR);
		return m_regFunc;
	}
};
typedef boost::shared_ptr<XMLvignetting> PtrXMLvignetting;

class XMLfuziness: protected XMLbase
{
	bool m_enabled;
	PtrFuncFtoF m_func;

public:
	XMLfuziness (string camera, string version, xercesc::DOMNode* node);
	
	bool isEnabled ()
	{
		return m_enabled;
	}

	PtrFuncFtoF getFunc ()
	{
		return m_func;
	}
};
typedef boost::shared_ptr<XMLfuziness> PtrXMLfuziness;


class XMLlensflare: protected XMLbase
{
public:
	bool m_enabled;
	uint2 m_sizeofenvmap;
	float2 m_envmapangle;
	float2 m_fovangle;

	float m_starMult, m_csMult1, m_csMult2;

public:
	XMLlensflare (string camera, string version, xercesc::DOMNode* node);
	
};
typedef boost::shared_ptr<XMLlensflare> PtrXMLlensflare;


class XMLmotionblur: protected XMLbase
{
	bool m_enabledParameters;
public:
	float3 deltaCameraPos;
	float targetDist;
	float2 deltaTarget;
	float2 fovAngle;

	uint numIterations;
public:
	XMLmotionblur (string camera, string version, xercesc::DOMNode* node);

	bool isEnabledParameters ()
	{
		return m_enabledParameters;
	}
/*	float3 getDeltapos ()
	{
		return m_deltapos;
	}
	float2 getDeltatarget ()
	{
		return m_deltatarget;
	}
	float getDistancetarget ()
	{
		return m_distancetarget;
	}
	uint getNumIterations ()
	{
		return m_numIterations;
	}*/
};
typedef boost::shared_ptr<XMLmotionblur> PtrXMLmotionblur;

class XMLdepthoffield: protected XMLbase
{
public:
	enum etype {TYPE_GATHER, TYPE_DIFFUSION};
public:
	etype m_type;
	int m_numIterations;

public:
	bool m_enabled;

	bool m_apertureIsFixed;
	float m_apertureValue;
	
	float m_sensorSize;
	float m_focalLen;
	float m_focusDist;

public:
	XMLdepthoffield (string camera, string version, xercesc::DOMNode* node);
};

typedef boost::shared_ptr<XMLdepthoffield> PtrXMLdepthoffield;


class XMLaperture: protected XMLbase
{
public:
	enum etype {TYPE_FIXED, TYPE_ADAPT};
private:
	etype m_type;
	float m_fnum;
	bool m_brightness;
	AdaptationParam m_param;
public:
	XMLaperture (string camera, string version, xercesc::DOMNode* node);

	etype getType ()
	{
		return m_type;
	}

	float getFnum ()
	{
		return m_fnum;
	}

	bool isEnabledBrightness ()
	{
		return m_brightness;
	}

	void getAdaptationParam (AdaptationParam& param)
	{
		param = m_param;
	}
};

typedef boost::shared_ptr<XMLaperture> PtrXMLaperture;

class XMLcolorfilterarray: protected XMLbase
{
private:
	FeatureColorFilterArrayParam m_red, m_green, m_blue;
	bool m_enabled;
public:
	XMLcolorfilterarray (string camera, string version, xercesc::DOMNode* node);

	bool isEnabled ()
	{
		return m_enabled;
	}

	void getRed (FeatureColorFilterArrayParam& param)
	{
		param = m_red;
	}

	void getGreen (FeatureColorFilterArrayParam& param)
	{
		param = m_green;
	}

	void getBlue (FeatureColorFilterArrayParam& param)
	{
		param = m_blue;
	}
};

typedef boost::shared_ptr<XMLcolorfilterarray> PtrXMLcolorfilterarray;


class XMLcamera: protected XMLbase
{
	PtrXMLvignetting m_vignetting;
	PtrXMLgeometricDistortion m_distortion;
	PtrXMLchromAbberation m_chromabberation;
	PtrXMLfuziness m_fuziness;
	PtrXMLlensflare m_lensflare;
	PtrXMLmotionblur m_motionblur;
	PtrXMLdepthoffield m_depthoffield;
	PtrXMLaperture m_aperture;
	PtrXMLcolorfilterarray m_cfa;

public:
	PtrXMLfunc m_funcIrrToSignal;
	bool m_isNoiseEnabled;
	PtrXMLfunc m_funcSignalToNoise;
	float m_noiseMin, m_noiseMax;

	enum efeature {NONE, DOF, BLUR, MOTION, LENSFLARE, DIST, CHROMABBER,ADAPTDO,
		ADAPTGET, CLAMP,IRRTOSIGNAL, NOISE, CFA, VIGNETTING};
	std::vector<efeature> m_vecFeatures;

public:
	XMLcamera (string camera, string version, xercesc::DOMNode* node);

	PtrXMLvignetting getVignetting ()
	{
		return m_vignetting;
	}

	PtrXMLchromAbberation getChromAbberation ()
	{
		return m_chromabberation;
	}

	PtrXMLgeometricDistortion getDistortion ()
	{
		return m_distortion;
	}

	PtrXMLfuziness getFuziness ()
	{
		return m_fuziness;
	}
	PtrXMLlensflare getLensflare ()
	{
		return m_lensflare;
	}
	PtrXMLmotionblur getMotionblur ()
	{
		return m_motionblur;
	}

	PtrXMLdepthoffield getDepthOfField ()
	{
		return m_depthoffield;
	}

	PtrXMLaperture getAperture ()
	{
		return m_aperture;
	}

	PtrXMLcolorfilterarray getColorFilterArray ()
	{
		return m_cfa;
	}
};


class XMLinputFile: public XMLbase
{
public:
	string path, red, green, blue, depth, envred, envgreen, envblue;

	enum etype {TYPE_STANDARD, TYPE_EXR};

	etype type;

	XMLinputFile (xercesc::DOMNode* node):
		XMLbase ("", "", "input")
	{
		path = XMLbase::getAttributeString (node, "path");
		
		string strtype = XMLbase::getAttributeString (node, "type");
		if (strtype == "exr") 
			type = TYPE_EXR;
		else if (strtype == "standard")
			type = TYPE_STANDARD;
		else
			throwInvalid ("type");

		for (int i = 0; i < 7; i++)
		{
			string in;
			string* out;
			switch (i)
			{
			case 0: in = "red"; out=&red; break;
			case 1: in = "green"; out=&green; break;
			case 2: in = "blue"; out=&blue; break;
			case 3: in = "depth"; out=&depth; break;
			case 4: in = "envred"; out=&envred; break;
			case 5: in = "envgreen"; out=&envgreen; break;
			case 6: in = "envblue"; out=&envblue; break;
			}
			try {
				*out = XMLbase::getAttributeString (node, in);
			} catch (XMLexception&) { }
		}
	}
	XMLinputFile (XMLinputFile& p):
		XMLbase ("", "", "input")
	{
		path = p.path;
		red = p.red;
		green = p.green;
		blue = p.blue;
		depth = p.depth;
		envred = p.envred;
		envgreen = p.envgreen;
		envblue = p.envblue;
		type = p.type;
	}
};


class XMLvideoExt: public XMLbase
{
public:
	int start, end, delay, loop;

	XMLvideoExt (xercesc::DOMNode* node);

	XMLvideoExt (): 
		XMLbase ("", "", "")
	{
	}

	void cloneMeTo (XMLvideoExt* p)
	{
		p->start = start;
		p->end = end;
		p->delay = delay;
		p->loop = loop;
	}
};

typedef boost::shared_ptr<XMLvideoExt> PtrXMLvideoExt;


class XMLinput;
typedef boost::shared_ptr<XMLinput> PtrXMLinput;

class XMLinput: public XMLbase
{
public:
	PtrXMLvideoExt videoExt;

	std::vector<XMLinputFile> files;
	float colorMult;
	bool depthLinearize;
	float depthMult, depthAdd, depthN, depthF;
	float icosmult, icosanglemult;

	XMLinput (xercesc::DOMNode* node);
	//PtrXMLinput clone ();
	XMLinput (): XMLbase ("", "", "") {}

	void cloneMeTo (XMLinput* p)
	{
		p->files = files;
		p->colorMult = colorMult;
		p->depthLinearize = depthLinearize;
		p->depthAdd = depthAdd;
		p->depthMult = depthMult;
		p->depthN = depthN;
		p->depthF = depthF;
		p->icosmult = icosmult;
		p->icosanglemult = icosanglemult;

		PtrXMLvideoExt ext = PtrXMLvideoExt (new XMLvideoExt);
		videoExt->cloneMeTo (ext.get ());
		p->videoExt = ext;
	}

	PtrXMLinput getSequenceOfVideo (int numIter)
	{
		PtrXMLinput input = PtrXMLinput (new XMLinput());
		cloneMeTo (input.get());

		int value = numIter + videoExt->start;
		
		for (size_t i = 0; i < input->files.size(); i++)
		{
			std::string s = input->files[i].path;
			char buffer [1024];
			sprintf (buffer, s.c_str(), value);
			input->files[i].path = buffer;
		}
		return input;
	}
};


class XMLoutputDemo: public XMLbase
{
public:
	XMLoutputDemo (xercesc::DOMNode* node);
	
	bool lightEnabled;
	float lightMult;
	string lightFilepath;

	uint2 wndSize;

};
typedef boost::shared_ptr<XMLoutputDemo> PtrXMLoutputDemo;



class XMLoutputFile: public XMLbase
{
public:
	XMLoutputFile (xercesc::DOMNode* node);

	string fileType, filePrefix, fileSuffix;
	uint fileNumDigits;
	uint itBegin, itStep, itCount;

	uint2 wndSize;
};
typedef boost::shared_ptr<XMLoutputFile> PtrXMLoutputFile;


class XMLoutput: public XMLbase
{
	PtrXMLoutputDemo m_demo;
	PtrXMLoutputFile m_file;
public:
	XMLoutput (xercesc::DOMNode* node);
	enum etype {TYPE_DEMO, TYPE_FILE} type;

	PtrXMLoutputDemo getDemo ()
	{
		if (type != TYPE_DEMO)
			throwException ("internal error");
		return m_demo;
	}

	PtrXMLoutputFile getFile ()
	{
		if (type != TYPE_FILE)
			throwException ("internal error");
		return m_file;
	}
};
typedef boost::shared_ptr<XMLoutput> PtrXMLoutput;



class XMLparser: protected XMLbase
{
	boost::shared_ptr<XMLcamera> m_camera;

	PtrXMLinput m_input;
	PtrXMLoutput m_output;

public:
	XMLparser (string filepath, string camera, string version, string input, string output);
	boost::shared_ptr<XMLcamera> getCamera ()
	{
		return m_camera;
	}
	PtrXMLinput getInput ()
	{
		return m_input;
	}
	PtrXMLoutput getOutput ()
	{
		return m_output;
	}
};